#ifndef __H2OPENGL__
#define __H2OPENGL__

#include "dd_data.h"

enum {VX, VY, VZ};	// Vertex indices.

// Lumobj Flags.
#define LUMF_USED		0x1
#define LUMF_RENDERED	0x2

typedef struct lumobj_s// For dynamic lighting.
{
	struct lumobj_s *next;	// Next in the same DL block, or NULL.
	int		flags;
	mobj_t	*thing;
	float	top, height;
	int		patch, radius, distance;
	byte	rgb[3];	// The color.
} lumobj_t;

// ScreenBits is currently unused.
extern int screenWidth, screenHeight, screenBits;
extern int viewph, viewpw, viewpx, viewpy;

void I_InitGraphics(void);
void I_ShutdownGraphics(void);

int I_ChangeResolution(int w, int h);

void OGL_InitRenderer(void);
void OGL_InitData(void);
void OGL_ResetData(void);

// Returns a pointer to a copy of the screen. The pointer must be 
// deallocated by the caller.
unsigned char *OGL_GrabScreen(void);

void OGL_SwitchTo3DState();
void OGL_Restore2DState(int step);	// Step 1: matrices, 2: attributes.
void OGL_UseWhiteFog(int yes);

float PointDist2D(float c[2]);
int thingDLRadius(lumobj_t *lum);

void R_RenderSprite(vissprite_t *spr);

// 2D drawing routines.
void OGL_DrawPatch_CS(int x, int y, int lumpnum);
void OGL_DrawPatch(int x, int y, int lumpnum);
void OGL_DrawPatchLitAlpha(int x, int y, float light, float alpha, int lumpnum);
void OGL_DrawFuzzPatch(int x, int y, int lumpnum);
void OGL_DrawAltFuzzPatch(int x, int y, int lumpnum);
void OGL_DrawShadowedPatch(int x, int y, int lumpnum);
void OGL_DrawRawScreen(int lump);	// Raw screens are 320 x 200.
void OGL_DrawLine(float x1, float y1, float x2, float y2, 
				  float r, float g, float b, float a);
void OGL_DrawRect(float x, float y, float w, float h, float r, float g, float b, float a);
void OGL_DrawRectTiled(int x, int y, int w, int h, int tw, int th);
void OGL_DrawCutRectTiled(int x, int y, int w, int h, int tw, int th, 
						  int cx, int cy, int cw, int ch);
void OGL_SetColor(int palidx);
void OGL_SetColorAndAlpha(float r, float g, float b, float a);
void OGL_DrawPSprite(int x, int y, float scale, int flip, int lump);

// Filters.
void OGL_SetFilter(int filterRGBA);
int OGL_DrawFilter();


// ogl_tex.c
typedef struct
{
	unsigned short	w, h;
	short			offx, offy;
	unsigned short	w2;		// For split textures, width of the other part.
} texsize_t;

extern texsize_t *lumptexsizes;	// Sizes for all the lumps. 
extern unsigned short *spriteheights;

extern float		texw, texh;
extern int			texmask;	
extern unsigned int	curtex;
extern int			pallump;

int OGL_InitPalettedTexture();
void OGL_TexInit();
void OGL_TexReset(void);
void OGL_ClearTextureMemory(void);
void OGL_TexDestroy();
void OGL_ResetLumpTexData();
void OGL_UpdateGamma(void);

void PalToRGB(byte *palidx, byte *rgb);
void PalIdxToRGB(byte *pal, int idx, byte *rgb);
unsigned int OGL_BindTexFlat(int lump);
void OGL_SetFlat(int idx);
void OGL_BindTexture(DGLuint texname);
void OGL_TextureFilterMode(int target, int parm);

void OGL_LowRes();

// Returns the DGL texture name.
DGLuint OGL_PrepareTexture(int idx);
DGLuint OGL_PrepareFlat(int idx);// Returns the OpenGL name of the texture.
DGLuint OGL_PrepareLightTexture();	// The dynamic light map.
DGLuint OGL_PrepareFlareTexture(int flare);

void OGL_SetTexture(int idx);
unsigned int OGL_PrepareSky(int idx, boolean zeroMask);
void OGL_GetSkyTopColor(int texidx, byte *rgb);

void OGL_SetSprite(int pnum);
unsigned int OGL_PrepareSprite(int pnum);
void OGL_SetTranslatedSprite(int pnum, unsigned char *trans);
void OGL_GetSpriteColor(int pnum, unsigned char *rgb);

void OGL_NewSplitTex(int lump, DGLuint part2name);
DGLuint OGL_GetOtherPart(int lump);

// Part is either 1 or 2. Part 0 means only the left side is loaded.
// No splittex is created in that case. Once a raw image is loaded
// as part 0 it must be deleted before the other part is loaded at the
// next loading.
void OGL_SetRawImage(int lump, int part);
void OGL_SetPatch(int lump);	// No mipmaps are generated.
void OGL_SetNoTexture();

int OGL_GetLumpTexWidth(int lump);
int OGL_GetLumpTexHeight(int lump);
int OGL_ValidTexHeight2(int width, int height);

void OGL_UpdateTexParams(int mipmode);
void OGL_UpdateRawScreenParams(int smoothing);

// returns the real DGL texture, if such exists
unsigned int OGL_GetTextureName(int texidx); 

// Only for textures (not for flats, sprites, etc.)
void OGL_DeleteTexture(int texidx); 


#include "m_bams.h"

// ogl_clip.c
typedef struct clipnode_s
{
	int					used;			// 1 if the node is in use.
	struct clipnode_s	*prev, *next;	// Previous and next nodes.
	binangle_t			start, end;		// The start and end angles (start < end).
} clipnode_t;

extern clipnode_t *clipnodes;	// The list of clipnodes.
extern clipnode_t *cliphead;	// The head node.

void C_Init();
void C_ClearRanges();
void C_Ranger();
void C_SafeAddRange(binangle_t startAngle, binangle_t endAngle);

// Add a segment relative to the current viewpoint.
void C_AddViewRelSeg(float x1, float y1, float x2, float y2);

// Check a segment relative to the current viewpoint.
int C_CheckViewRelSeg(float x1, float y1, float x2, float y2);

// Returns 1 if the specified angle is visible.
int C_IsAngleVisible(binangle_t bang);

clipnode_t *C_AngleClippedBy(binangle_t bang);

// Returns 1 if the subsector might be visible.
int C_CheckSubsector(subsector_t *ssec);


// ogl_sky.c

#include "gl_sky.h"

// ogl_halo.c

void H_Clear();

// Prepares the list of halos to render. Goes through the luminousList.
void H_InitForNewFrame();

// This must be called when the renderer is in player sprite rendering mode:
// 2D projection to the game view. 
void H_Render();


#endif